import json
import os
import re
import click

from rc3.common import json_helper, print_helper, helper_functions, decorators

PATTERN = re.compile(r'{{(.*?)}}')
JSON_FILE_PATTERN = re.compile(r'"{{(.*?#file.*?)}}"')


def lookup_one_var(var):
    envs = [
        json_helper.read_environment('current')[1],
        json_helper.read_environment('global')[1],
        os.environ
    ]
    return lookup_var_value(envs, var)


def process_subs(wrapper):
    if has_vars(wrapper):
        sub_vars(wrapper)
    # print_helper.print_json(r.get('_original', None))


def sub_vars(wrapper):
    envs = [
        json_helper.read_environment('current')[1],
        json_helper.read_environment('global')[1],
        os.environ
    ]
    r = wrapper.get('_original')

    # sub dicts
    sub_in_dict(envs, r.get('form_data'))
    sub_in_dict(envs, r.get('headers'))
    sub_in_dict(envs, r.get('params'))
    sub_in_dict(envs, r.get('auth'))

    # sub strings (& json body)
    r['url'] = sub_in_string(envs, r.get('url'))
    text = r.get('body', {}).get('text')
    if text is not None:
        r.get('body')['text'] = sub_in_string(envs, text)
    _json = r.get('body', {}).get('json')
    if _json is not None:
        json_string = json.dumps(_json)
        # special sub for {{ #file }} in JUST the body.json portion of the .request
        new_string = sub_file_in_json_string(envs, json_string)
        new_string = sub_in_string(envs, new_string)
        if new_string != json_string:
            r.get('body')['json'] = json.loads(new_string)


def lookup_var_value(envs, var, seen=None):
    # if var starts with #, then treat is as a helper function and NOT env var
    if var.startswith("#"):
        return helper_functions.lookup_helper_value(var)

    # seen just holds vars that have already seen for THIS lookup
    # if already seen, then we have an infinite loop...
    if seen is None:
        seen = []
    if var in seen:
        raise click.ClickException(
            f'var {{{{{var}}}}} has caused an infinite loop during lookup, please check/update your vars!')
    else:
        seen.append(var)

    for env in envs:
        if var in env:
            outer_value = env.get(var)
            for match in PATTERN.finditer(outer_value):
                inner_var = match.group(1).strip()
                inner_value = lookup_var_value(envs, inner_var, seen)
                outer_value = outer_value.replace(match.group(0), inner_value)
            return outer_value
    raise click.ClickException(
        f'var {{{{{var}}}}} is in the REQUEST but cannot be found in the current, global, or OS environment')


def sub_in_dict(envs, d):
    if d is None:
        return
    # pattern = re.compile(r'{{(.*?)}}')
    for key, value in d.items():
        new_value = value
        for match in PATTERN.finditer(value):
            var = match.group(1).strip()
            var_value = lookup_var_value(envs, var)
            # this allows multiple vars to be used in a single value (each gets replaced)
            new_value = new_value.replace(match.group(0), var_value)
        d[key] = new_value


def sub_file_in_json_string(envs, s):
    if s is None:
        return None
    # JUST for {{ #file }} expressions in json.body, we want to replace surrounding double quotes also!
    # I.E.: "{{ #file }}" and not just {{ #file }}
    # NOTE: if #file is used in other parts of the request, it will do normal subbing, and not replace ""
    for match in JSON_FILE_PATTERN.finditer(s):
        var = match.group(1).strip()
        var_value = lookup_var_value(envs, var)
        s = s.replace(match.group(0), var_value)
    return s


def sub_in_string(envs, s):
    if s is None:
        return None
    # pattern = re.compile(r'{{(.*?)}}')
    for match in PATTERN.finditer(s):
        var = match.group(1).strip()
        var_value = lookup_var_value(envs, var)
        s = s.replace(match.group(0), var_value)
    return s


def has_vars(wrapper):
    r = wrapper.get('_original')
    dicts = [
        r.get('form_data', {}),
        r.get('headers', {}),
        r.get('params', {}),
        r.get('auth', {})
    ]
    strings = [
        r.get('url', ''),
        r.get('body', {}).get('text', ''),
        json.dumps(r.get('body', {}).get('json', {}))
    ]
    for d in dicts:
        for v in d.values():
            strings.append(v)

    # pattern = re.compile(r'{{(.*?)}}')
    for s in strings:
        match = PATTERN.search(s)
        if match is not None:
            return True
    return False



import json
import os
import re
import uuid

import click
import pkce
from click import prompt

from rc3.common import json_helper, print_helper, decorators, file_helper


def lookup_helper_value(var):
    # simple hard-coding for now
    # in the future maybe dynamically use any function in this file, or even add user-defined helper functions
    if var.startswith("#pkce_cvcc"):
        return pkce_cvcc(var)
    if var.startswith("#uuid"):
        return uuid_helper(var)
    if var.startswith("#prompt"):
        return prompt_helper(var)
    if var.startswith("#file"):
        return file_helper_function(var)
    raise click.ClickException(
        f'handlebar helper_function [{var}] is invalid!')


def parse_env_var(var):
    parts = var.split()
    helper_name = parts[0]
    var_name = None
    env_name = "global"
    if len(parts) > 1:
        var_name = parts[1]
        var_parts = var_name.split('.')
        if len(var_parts) > 1:
            env_name = var_parts[0]
            var_name = var_parts[1]
            if env_name not in ['global', 'current']:
                raise click.ClickException(
                    f'Env name in {helper_name} helper function must be global or current. [{env_name}] is invalid!')
    return env_name, var_name


def pkce_cvcc(var):
    # initial impl, uses default/mostly fixed values of
    # length = 128
    # global var to store = code_verifier
    # challenge transformation = S256
    parts = var.split()
    env_name, var_name = parse_env_var(var)
    if var_name is None:
        var_name = 'code_verifier'
    if len(parts) > 2:
        raise click.ClickException(
            f'Invalid # of parameters to {parts[0]} helper function.  Expected 0 or 1, but got {len(parts)}!')

    # generate cv and cc
    cv, cc = pkce.generate_pkce_pair()

    # store cv into global env
    env_filename, env = json_helper.read_environment(env_name)
    env[var_name] = cv
    json_helper.write_environment(env_filename, env)

    # bust the cache, so future reads in same process see the change
    decorators.rc_clear_cache('read_environment')

    # return cc, to be populated in template
    return cc


def uuid_helper(var):
    # initial impl
    # type = uuid4
    # no parameter = NO var
    # parameter with no . = stored in global
    # parameter with . = LEFT side must be global|current RIGHT side is var name
    parts = var.split()
    env_name, var_name = parse_env_var(var)
    if len(parts) > 2:
        raise click.ClickException(
            f'Invalid # of parameters to #uuid helper function.  Expected 0 or 1, but got {len(parts)}!')

    # generate
    value = str(uuid.uuid4())

    # store in env
    if var_name is not None:
        env_filename, env = json_helper.read_environment(env_name)
        env[var_name] = value
        json_helper.write_environment(env_filename, env)

    # bust the cache, so future reads in same process see the change
    decorators.rc_clear_cache('read_environment')

    # return the uuid
    return value


def prompt_helper(var):
    parts = var.split()
    if len(parts) < 2:
        raise click.ClickException('A prompt is required when using the #prompt helper function!')

    # look for a prompt + default
    p = var[8:]
    parts = p.split(":")
    default = ""
    if len(parts) > 1:
        p = parts[0]
        default = parts[1]

    # prompt for value, and return
    return click.prompt(p, default=default)


def file_helper_function(var):
    # IMPORTANT! preprocess_file_option() MUST have been called back in cmd_send BEFORE env subs/helpers are called!
    parts = var.split()
    if len(parts) > 1:
        raise click.ClickException('The #file helper function doesn\'t support any parameters!')
    if not file_helper.state['has_file']:
        raise click.ClickException('The --file option must be used since this request has a #file helper function!')

    # ALWAYS return as a string
    # if subbing into the JSON body, the string will get converted back to JSON/dict after subs
    return file_helper.consume_as_string()






import os
import shutil

import click
from rc3.common import config_helper, json_helper, data_helper
from rc3.common.data_helper import SETTINGS_FILENAME, COLLECTION_FILENAME, GLOBAL_ENV_FILENAME


@click.command("new", short_help="Create a new Collection in CWD")
def cli():
    """\b
    Must be run from an empty directory.
    Will:
    1. Create a new rc-collection.json
    2. Optionally create example Requests
    3. Import your new collection into RC_HOME/rc-settings.json
    """

    empty_cwd_check()
    init_cwd_as_collection()
    import_collection()


def empty_cwd_check():
    cwd = os.getcwd()
    if os.listdir(cwd):
        print("CWD must be empty to create a new collection.")
        dest = os.path.join(cwd, COLLECTION_FILENAME)
        if os.path.exists(dest):
            print("Try 'rc import' to import an existing collection...")
        raise click.Abort()


def init_cwd_as_collection():
    cwd = os.getcwd()
    dir_name = os.path.basename(cwd)
    name = click.prompt("Enter a NAME for this COLLECTION", default=dir_name)
    include_examples = click.confirm("Include example Requests in your new collection?", default=True, show_default=True)
    if include_examples:
        # copy the entire example collection
        data_helper.copy_tree('collection', cwd)
    else:
        # copy JUST environments and the rc-collection file
        env_dir = os.path.join(cwd, 'environments')
        data_helper.copy_tree('collection/environments', env_dir)
        data_helper.copy('collection/rc-collection.json', cwd)

    c, wrapper = json_helper.read_collection(_dir=cwd)
    c['name'] = name
    if not include_examples:
        c['current_request'] = ""
    json_helper.write_collection(wrapper)


def import_collection():
    cwd = os.getcwd()
    print("Importing collection into RC_HOME/rc-settings.json")
    collection_dict = json_helper.load_and_validate(COLLECTION_FILENAME, _dir=cwd)
    if collection_dict is None:
        return

    # get "name" from json, or abort
    name = collection_dict.get("name", None)
    if name is None:
        print("Error: rc-collection.json doesn't have a valid name attribute")
        raise click.Abort

    # for any .defaults environments in the collection, copy them to .json (i.e. make a real environment)
    # note: there should also be a .gitignore in the env folder, so .json doesn't get committed
    env_folder = os.path.join(cwd, 'environments')
    defaults_counter = 0
    for dirpath, dirnames, files in os.walk(env_folder):
        for file in files:
            if file.endswith('.defaults'):
                env_name = file.split('.')[-2]
                default_file = os.path.join(dirpath, file)
                new_file = os.path.join(dirpath, env_name + ".json")
                if not os.path.exists(new_file):
                    defaults_counter += 1
                    shutil.copy(default_file, new_file)
    # if defaults_counter > 0:
    #     print(f"{defaults_counter} default environment(s) initialized in your collection")

    settings = json_helper.read_settings()
    settings["current_collection"] = name

    # if collection name already exists, just update the location pointer
    # NOTE: the old collection location will disappear from settings...
    found = False
    for cp in settings['collections']:
        if cp['name'] == name:
            cp['location'] = cwd
            found = True
            break
    if not found:
        settings["collections"].append({
            "name": name,
            "location": cwd
        })
    json_helper.write_settings(settings)
    # print(f"Collection '{name}' has been imported successfully, try 'rc list' to see available Requests.")
    print(f"Collection '{name}' has been successfully imported, try 'rc list' next...")



import os
import shutil

import click

from rc3.commands import cmd_new
from rc3.common import config_helper, json_helper, data_helper
from rc3.common.data_helper import SETTINGS_FILENAME, COLLECTION_FILENAME, GLOBAL_ENV_FILENAME


@click.command("import", short_help="Import an existing Collection from CWD")
def cli():
    """\b
    Must be run from a directory with a valid rc-collection.json.
    Will:
    1. Import the collection into RC_HOME/rc-settings.json
    """

    valid_collection_check()
    cmd_new.import_collection()


def valid_collection_check():
    cwd = os.getcwd()
    dest = os.path.join(cwd, COLLECTION_FILENAME)
    if not os.path.exists(dest):
        print("CWD must contain a valid rc-collection.json file")
        raise click.Abort()



import json
import os
import time

import click

from rc3.commands import cmd_list
from rc3.common import json_helper, print_helper, config_helper
from rc3.common.data_helper import SCHEMA_BASE_URL, SCHEMA_PREFIX, SCHEMA_VERSION


@click.command("upgrade", short_help="Attempt to upgrade the current COLLECTION.")
def cli():
    """\b
    Upgrade the current collection where possible to your current version of the rc CLI.

    """
    show_fake_progress()


def show_fake_progress():
    print("Checking for possible upgrades...")
    with click.progressbar(range(16)) as bar:
        for x in bar:
            time.sleep(.25)
    click.confirm("Would you like to upgrade the EXAMPLE REQUESTS in this collection", default=True)
    print("Done.")
    click.confirm("Would you like to upgrade REQUEST SCHEMA references to the latest 0.0.9 in this collection", default=True)
    print("Done.")
    click.confirm("Would you like to upgrade REQUEST extract[*] format in this collection", default=True)
    print("Done.")



import os
import re

from rc3.common import data_helper
from rc3.common.data_helper import SETTINGS_FILENAME, GLOBAL_ENV_FILENAME
from rc3.common.decorators import rc_print_durations

DEFAULT_CONFIG_FOLDER = '.rc'
RC_HOME = 'RC_HOME'


def get_config_folder():
    rc_home = os.getenv(RC_HOME)
    if rc_home is None:
        home = os.path.expanduser("~")
        rc_home = os.path.join(home, DEFAULT_CONFIG_FOLDER)
    if not os.path.exists(rc_home):
        os.mkdir(rc_home)
    return rc_home


def clean_filename(name):
    if name is None:
        return None
    name = name.lower()
    name = re.sub(' ', '_', name)
    name = re.sub('[^a-zA-Z0-9_-]', '', name)
    return name


@rc_print_durations
def init_rc_home():
    # Note: this next cmd will always create RC_HOME / ~/.rc if it doesn't exist
    home = get_config_folder()

    dest = os.path.join(home, SETTINGS_FILENAME)
    if not os.path.exists(dest):
        # print("Creating " + dest)
        data_helper.copy('home/' + SETTINGS_FILENAME, dest)

    dest = os.path.join(home, GLOBAL_ENV_FILENAME)
    if not os.path.exists(dest):
        # print("Creating " + dest)
        data_helper.copy('home/' + GLOBAL_ENV_FILENAME, dest)

    dest = os.path.join(home, 'schemas')
    if not os.path.exists(dest):
        # print("Creating " + dest)
        data_helper.copy_tree('home/schemas', dest)



import json
from json import JSONDecodeError

import click

state = {
    'has_file': False,
    'consumed': False,
    '_json': None,
    '_text': None
}


def preprocess_file_option(file):
    if file is not None:
        state['has_file'] = True
        state['_json'] = read_as_json(file)
        if state['_json'] is None:
            state['_text'] = read_as_text(file)


def read_as_json(file):
    try:
        with click.open_file(file) as f:
            return json.load(f)
    except JSONDecodeError as e:
        return None


def read_as_text(file):
    with click.open_file(file) as f:
        return f.read()


def consume_as_string():
    state['consumed'] = True
    if state['_json'] is not None:
        return json.dumps(state['_json'])
    return state['_text']



import json
import os
import time

import click
import jwt

from rc3.commands import cmd_list
from rc3.common import json_helper, print_helper, config_helper, env_helper
from rc3.common.data_helper import SCHEMA_BASE_URL, SCHEMA_PREFIX, SCHEMA_VERSION


@click.command("decode", short_help="Decode a JWT stored in an ENVIRONMENT.")
@click.argument('jwt_env_var', type=str, required=False)
def cli(jwt_env_var):
    """\b
    Will attempt to decode the JWT_ENV_VAR env var.
    If no JWT_ENV_VAR argument is passed will attempt to decode the 'token' env var.

    """
    var = 'token' if jwt_env_var is None else jwt_env_var
    value = env_helper.lookup_one_var(var)
    if value is None:
        raise click.ClickException(f"No ENV VAR found for [{var}]")

    try:
        headers = jwt.get_unverified_header(value)
        payload = jwt.decode(value, options={"verify_signature": False})
    except jwt.exceptions.DecodeError as error:
        raise click.ClickException(f"There was an error decoding the JWT in ENV VAR [{var}]\n{error}")
        # print(error)

    print(f'Decoding HEADERS and CLAIMS from \'{var}\' env var')
    print_helper.print_json(headers)
    print_helper.print_json(payload)
    if 'iat' in payload:
        iat = payload['iat']
        print(f'Issued at:  {time.ctime(iat)}')
    if 'exp' in payload:
        exp = payload['exp']
        now = time.time()
        if exp < now:
            click.echo(click.style(f'Expired at: {time.ctime(exp)}', fg='red'))
            print(f'Curr time:  {time.ctime(time.time())}')
        else:
            print(f'Curr time:  {time.ctime(time.time())}')
            click.echo(click.style(f'Expires at: {time.ctime(exp)}', fg='green'))



import copy
import json
import os
import re
from json import JSONDecodeError

import click
import requests
from jsonpath_ng import parse
from requests.auth import HTTPBasicAuth, AuthBase
from requests.exceptions import SSLError

from rc3.commands import cmd_request
from rc3.common import json_helper, print_helper, env_helper, inherit_helper, rc_globals, file_helper, decorators


@click.command("send", short_help="Sends an HTTP request & writes results to a response file.")
@click.option('-p', '--pick', is_flag=True, default=False, help="Pick a REQUEST then send it.")
@click.option('-e', '--edit', is_flag=True, default=False, help="Edit a REQUEST then send it.")
@click.option('-f', '--file', type=click.Path(exists=True, readable=True, allow_dash=True, dir_okay=False), help="File to send as REQUEST body.")
@click.argument('request_name', type=str, required=False, nargs=-1)
def cli(pick, edit, file, request_name):
    """\b
    Will send an HTTP request using the Python requests library.
    \b
    REQUEST_NAME(s) is optional.
    REQUEST_NAME(s) will default to the current_request.
    REQUEST_NAME(s) if used should be one of:
    1. The NUM column from 'rc request --list' output
    2. THe NAME column from 'rc request --list' output

    \b
    OUTPUT will be the response body to STDOUT.
    OUTPUT will also be a *.response file in the same dir as the *.request file.
    """
    if len(request_name) > 1:
        if pick:
            raise click.ClickException("--pick option is invalid if multiple REQUEST_NAMES")
        if edit:
            raise click.ClickException("--edit option is invalid if multiple REQUEST_NAMES")
        if file is not None:
            raise click.ClickException("--file option is invalid if multiple REQUEST_NAMES")
        for name in request_name:
            r = lookup_request(name)
            # make a deepcopy, in case we are sending multiples of the same request, env var subs happen again!
            # & bust the read_env cache, so 2nd request can see extracted values from 1st request
            r = copy.deepcopy(r)
            decorators.rc_clear_cache('read_environment')
            send(r, None)
        return

    # otherwise 0 or 1 request_names
    name = None
    if len(request_name) == 1:
        name = request_name[0]

    if pick and edit:
        r = cmd_request.pick_request(name)
        r = cmd_request.edit_request(None, wrapper=r)
        send(r, file)
    elif pick:
        r = cmd_request.pick_request(name)
        send(r, file)
    elif edit:
        r = cmd_request.edit_request(name)
        send(r, file)
    else:
        r = lookup_request(name)
        send(r, file)


def lookup_and_send(request_name):
    r = lookup_request(request_name)
    send(r, None)


def lookup_request(request_name):
    r = cmd_request.lookup_request(request_name)
    if r is None and request_name is None:
        raise click.ClickException("There is no current REQUEST, exiting...")
    if r is None:
        raise click.ClickException("REQUEST '{}' not found. See 'rc request --list'".format(request_name))
    return r


def send(wrapper, file):
    settings = json_helper.read_settings()
    request = wrapper.get('_original')
    headers = request.get('headers', {})

    # find_auth() before var subs, so env var subs happen in inherited auth also!
    # IMPORTANT! preprocess_file_option() before env var subs, so {{#file}} helper can consume --file if necc.
    request['auth'] = inherit_helper.find_auth(wrapper)
    file_helper.preprocess_file_option(file)
    env_helper.process_subs(wrapper)

    # determine BODY to POST/PUT
    # potentially updates default headers
    # potentially overrides .request settings with --file passed in
    _json, _data = determine_body(request, file, headers)

    timeout = settings.get('request_timeout', 30)
    allow_redirects = settings.get('follow_redirects', False)
    verify = settings.get('ca_cert_verification', True)
    if verify and len(settings.get('ca_bundle','')) > 0:
        verify = settings.get('ca_bundle')
        if not os.path.exists(verify):
            raise click.ClickException(f'settings.json ca_bundle file [{verify}] does not exist!')
    if settings.get('headers_send_nocache', True):
        if 'Cache-Control' not in headers:
            headers['Cache-Control'] = 'no-cache'

    try:
        response = requests.request(request.get('method'), request.get('url'),
                                    headers=headers,
                                    json=_json,
                                    auth=create_auth(inherit_helper.find_auth(wrapper)),
                                    params=request.get('params', None),
                                    data=_data,
                                    timeout=timeout,
                                    allow_redirects=allow_redirects,
                                    verify=verify)
    except SSLError as e:
        print()
        print(type(e).__name__ + ": " + str(e))
        click.echo("SSLError: Try setting REQUESTS_CA_BUNDLE, CURL_CA_BUNDLE, or rc-settings.ca_bundle")
        click.echo("See: https://github.com/gswilcox01/rc3/tree/master?tab=readme-ov-file#ca-certificates")
        raise click.Abort

    process_output(wrapper, response)


def determine_body(request, file, headers):
    # --file option overrides all other definitions in .request file
    # IMPORTANT!: file_helper.preprocess_file_option(file) MUST have already been called before handlebar substitution!
    # IMP2: if the --file was consumed by a handlebar expression of "{{ #file }}" already, then it won't be used here!
    if file_helper.state['has_file'] and not file_helper.state['consumed']:
        return file_helper.state['_json'], file_helper.state['_text']

    # no --file option, error if multiple bodies defined
    _json = None
    _data = None
    body_count = 0
    _json = request.get('body', {}).get('json', None)
    if _json is not None:
        body_count += 1
    text = request.get('body', {}).get('text', None)
    if text is not None:
        body_count += 1
    form_data = request.get('form_data', None)
    if form_data is not None:
        body_count += 1
    if body_count > 1:
        raise click.ClickException("REQUEST can only have 1 of body.json, body.text, or form_data.  You have {}.".format(body_count))

    _data = None
    if text is not None:
        _data = text
        # don't overwrite header if explicitly defined in the .request
        if 'Content-Type' not in headers:
            headers['Content-Type'] = "text/plain"
    elif form_data is not None:
        _data = form_data
    return _json, _data


def process_output(wrapper, response):
    settings = json_helper.read_settings()
    cli_options = rc_globals.get_cli_options()
    request = wrapper.get('_original')

    # process potential extract, and update env file
    extract_buffer = process_extracts(request, response)
    verbose_output = extract_buffer['verbose_output']   # will also have extract_errors & extracted now

    # write out *.response file
    save_responses = request.get('save_responses', settings.get('save_responses', True))
    if save_responses:
        response_filename = os.path.join(wrapper.get('_dir'),
                                         wrapper.get('_filename').split('.')[0] + ".response")
        json_helper.write_json(response_filename, verbose_output)

    # if extracted to STDOUT that overrides everything
    # otherwise print verbose or just response.body
    if len(extract_buffer['stdout']) > 0:
        print_helper.print_json(extract_buffer['stdout'])
    elif cli_options.get('verbose', False):
        print_helper.print_json(verbose_output)
    else:
        print_helper.print_json_or_text(response)


def process_extracts(request, response):
    # env extracts will go straight to the env, but other "to"s and errors collected here
    verbose_output = create_verbose_output(response)
    verbose_output['extract_errors'] = []
    verbose_output['extracted'] = {}
    extract_buffer = {
        "stdout": {},
        "verbose_output": verbose_output,
        "response": verbose_output['extracted'],    # pointer here for easier ref
        "errors": verbose_output['extract_errors']  # pointer here for easier ref
    }

    # do nothing if no "extract" in the request
    extract_list = request.get("extract", None)
    if extract_list is None or len(extract_list) == 0:
        return extract_buffer

    for extract in extract_list:
        process_one_extract(extract, response, extract_buffer)
    return extract_buffer


def process_one_extract(extract, response, extract_buffer):
    to_environment = False
    to = extract.get("to", "global")
    if to in ["global", "current"]:
        to_environment = True
        env_filename, env = json_helper.read_environment(to)

    value = None
    if "json_path" in extract:
        value = extract_json_path(extract, response, extract_buffer)
    elif "text_pattern" in extract:
        value = extract_regex(extract, response, extract_buffer)

    # write to environment if that is the output
    # otherwise write to the buffer where we are collecting 1..n extracts
    var = extract.get("var", "token")
    # print(f'var={var} value={value} to={to}')
    if to_environment:
        env[var] = value
        json_helper.write_environment(env_filename, env)
    elif to == "stdout":
        extract_buffer.get("stdout")[var] = value
    elif to == "response":
        extract_buffer.get("response")[var] = value
    else:
        raise click.ClickException(f"Unrecognized $.extract.to ({to}), please check your .request")


def extract_json_path(extract, response, extract_buffer):
    # attempt to find json path
    # See: https://www.digitalocean.com/community/tutorials/python-jsonpath-examples
    # See: https://jsonpath.com/
    # See: https://pypi.org/project/jsonpath-ng/
    json_path = extract.get("json_path", None)
    errors = extract_buffer['errors']

    # body=response.body, response=verbose_output
    extract_from = extract.get("from", "body")
    target = extract_body_as_json(response) if extract_from == "body" else extract_buffer['verbose_output']
    if target is None:
        errors.append("REQUEST has $.extract.json_path, but response is not JSON!")
        return

    jsonpath_expression = parse(json_path)
    match = jsonpath_expression.find(target)
    if len(match) < 1:
        errors.append("REQUEST json_path=[{}] had no matches in response JSON!".format(json_path))
        return
    # use the value from the first match
    return match[0].value


def extract_regex(extract, response, extract_buffer):
    # See python regex docs here for valid regex patterns:
    # https://docs.python.org/3/howto/regex.html
    text_pattern = extract.get("text_pattern", None)
    errors = extract_buffer['errors']

    # body=response.body, response=verbose_output
    extract_from = extract.get("from", "body")
    target = response.text if extract_from == "body" else print_helper.get_json_string(extract_buffer['verbose_output'])
    if len(target) < 1 and extract_from == "body":
        errors.append("REQUEST has $.extract.text_pattern, but response.text is 0 length!")
        return
    if "(" not in text_pattern or ")" not in text_pattern:
        errors.append("REQUEST $.extract.text_pattern, MUST contain a matching group ()!")
        return

    # do it
    pattern = re.compile(text_pattern)
    match = pattern.search(target)
    # group(0) is the WHOLE match, (1) is the first group in the pattern (what we want)
    if match is None:
        errors.append("REQUEST text_pattern=[{}] had no matches in response JSON!".format(text_pattern))
        return
    return match.group(1)


def extract_body_as_json(response):
    try:
        _json = response.json()
        return _json
    except BaseException as error:
        return None


def create_verbose_output(response):
    body = {}
    if extract_body_as_json(response) is None:
        body['text'] = response.text
    else:
        body['json'] = extract_body_as_json(response)

    headers = {}
    header_size = 0
    for key, value in response.headers.items():
        header_size += len(key) + len(value)
        headers[key.lower()] = value

    # verbose output
    return {
        "status_code": response.status_code,
        "time": str(divmod(response.elapsed.microseconds, 1000)[0]) + "ms",
        # "time2": str(response.elapsed.microseconds / 1000) + "ms",
        "size": {
            "body": len(response.content),
            "headers": header_size,
            "total": len(response.content) + header_size
        },
        "headers": headers,
        "body": body
    }


class HTTPTokenAuth(AuthBase):
    def __init__(self, token, token_header="Authorization", token_name="Bearer"):
        self.token = token
        self.token_header = token_header
        self.token_name = token_name

    def __call__(self, r):
        value = self.token_name + " " + self.token
        r.headers[self.token_header] = value.strip()
        return r


def create_auth(auth_config):
    _auth = None
    _type = auth_config.get('type')
    if _type == 'basic':
        _auth = HTTPBasicAuth(auth_config.get('username'), auth_config.get('password'))
    if _type == 'bearer':
        _auth = HTTPTokenAuth(auth_config.get('bearer_token'))
    if _type == 'token':
        _auth = HTTPTokenAuth(auth_config.get('token_value'),
                              token_name="",
                              token_header=auth_config.get('token_header', "Authorization"))
    return _auth
